@@ -1,19 +1,19 @@
-/* A Bison parser, made by GNU Bison 2.6.4.  */
+/* A Bison parser, made by GNU Bison 3.0.2.  */
 
 /* Bison interface for Yacc-like parsers in C
-   
-      Copyright (C) 1984, 1989-1990, 2000-2012 Free Software Foundation, Inc.
-   
+
+   Copyright (C) 1984, 1989-1990, 2000-2013 Free Software Foundation, Inc.
+
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
@@ -26,13 +26,13 @@
    special exception, which will cause the skeleton and the resulting
    Bison output files to be licensed under the GNU General Public
    License without this special exception.
-   
+
    This special exception was added by the Free Software Foundation in
    version 2.2 of Bison.  */
 
 #ifndef YY__XKBCOMMON_SRC_XKBCOMP_PARSER_H_INCLUDED
 # define YY__XKBCOMMON_SRC_XKBCOMP_PARSER_H_INCLUDED
-/* Enabling traces.  */
+/* Debug traces.  */
 #ifndef YYDEBUG
 # define YYDEBUG 0
 #endif
@@ -40,76 +40,75 @@
 extern int _xkbcommon_debug;
 #endif
 
-/* Tokens.  */
+/* Token type.  */
 #ifndef YYTOKENTYPE
 # define YYTOKENTYPE
-   /* Put the tokens into the symbol table, so that GDB and other debuggers
-      know about them.  */
-   enum yytokentype {
-     END_OF_FILE = 0,
-     ERROR_TOK = 255,
-     XKB_KEYMAP = 1,
-     XKB_KEYCODES = 2,
-     XKB_TYPES = 3,
-     XKB_SYMBOLS = 4,
-     XKB_COMPATMAP = 5,
-     XKB_GEOMETRY = 6,
-     XKB_SEMANTICS = 7,
-     XKB_LAYOUT = 8,
-     INCLUDE = 10,
-     OVERRIDE = 11,
-     AUGMENT = 12,
-     REPLACE = 13,
-     ALTERNATE = 14,
-     VIRTUAL_MODS = 20,
-     TYPE = 21,
-     INTERPRET = 22,
-     ACTION_TOK = 23,
-     KEY = 24,
-     ALIAS = 25,
-     GROUP = 26,
-     MODIFIER_MAP = 27,
-     INDICATOR = 28,
-     SHAPE = 29,
-     KEYS = 30,
-     ROW = 31,
-     SECTION = 32,
-     OVERLAY = 33,
-     TEXT = 34,
-     OUTLINE = 35,
-     SOLID = 36,
-     LOGO = 37,
-     VIRTUAL = 38,
-     EQUALS = 40,
-     PLUS = 41,
-     MINUS = 42,
-     DIVIDE = 43,
-     TIMES = 44,
-     OBRACE = 45,
-     CBRACE = 46,
-     OPAREN = 47,
-     CPAREN = 48,
-     OBRACKET = 49,
-     CBRACKET = 50,
-     DOT = 51,
-     COMMA = 52,
-     SEMI = 53,
-     EXCLAM = 54,
-     INVERT = 55,
-     STRING = 60,
-     INTEGER = 61,
-     FLOAT = 62,
-     IDENT = 63,
-     KEYNAME = 64,
-     PARTIAL = 70,
-     DEFAULT = 71,
-     HIDDEN = 72,
-     ALPHANUMERIC_KEYS = 73,
-     MODIFIER_KEYS = 74,
-     KEYPAD_KEYS = 75,
-     FUNCTION_KEYS = 76,
-     ALTERNATE_GROUP = 77
-   };
+  enum yytokentype
+  {
+    END_OF_FILE = 0,
+    ERROR_TOK = 255,
+    XKB_KEYMAP = 1,
+    XKB_KEYCODES = 2,
+    XKB_TYPES = 3,
+    XKB_SYMBOLS = 4,
+    XKB_COMPATMAP = 5,
+    XKB_GEOMETRY = 6,
+    XKB_SEMANTICS = 7,
+    XKB_LAYOUT = 8,
+    INCLUDE = 10,
+    OVERRIDE = 11,
+    AUGMENT = 12,
+    REPLACE = 13,
+    ALTERNATE = 14,
+    VIRTUAL_MODS = 20,
+    TYPE = 21,
+    INTERPRET = 22,
+    ACTION_TOK = 23,
+    KEY = 24,
+    ALIAS = 25,
+    GROUP = 26,
+    MODIFIER_MAP = 27,
+    INDICATOR = 28,
+    SHAPE = 29,
+    KEYS = 30,
+    ROW = 31,
+    SECTION = 32,
+    OVERLAY = 33,
+    TEXT = 34,
+    OUTLINE = 35,
+    SOLID = 36,
+    LOGO = 37,
+    VIRTUAL = 38,
+    EQUALS = 40,
+    PLUS = 41,
+    MINUS = 42,
+    DIVIDE = 43,
+    TIMES = 44,
+    OBRACE = 45,
+    CBRACE = 46,
+    OPAREN = 47,
+    CPAREN = 48,
+    OBRACKET = 49,
+    CBRACKET = 50,
+    DOT = 51,
+    COMMA = 52,
+    SEMI = 53,
+    EXCLAM = 54,
+    INVERT = 55,
+    STRING = 60,
+    INTEGER = 61,
+    FLOAT = 62,
+    IDENT = 63,
+    KEYNAME = 64,
+    PARTIAL = 70,
+    DEFAULT = 71,
+    HIDDEN = 72,
+    ALPHANUMERIC_KEYS = 73,
+    MODIFIER_KEYS = 74,
+    KEYPAD_KEYS = 75,
+    FUNCTION_KEYS = 76,
+    ALTERNATE_GROUP = 77
+  };
 #endif
 /* Tokens.  */
 #define END_OF_FILE 0
@@ -176,22 +175,21 @@ extern int _xkbcommon_debug;
 #define FUNCTION_KEYS 76
 #define ALTERNATE_GROUP 77
 
-
-
+/* Value type.  */
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
-typedef union YYSTYPE
+typedef union YYSTYPE YYSTYPE;
+union YYSTYPE
 {
-/* Line 2077 of yacc.c  */
-#line 127 "../src/xkbcomp/parser.y"
+#line 162 "src/xkbcomp/parser.y" /* yacc.c:1909  */
 
         int              ival;
-        unsigned         uval;
         int64_t          num;
         enum xkb_file_type file_type;
         char            *str;
         xkb_atom_t      sval;
         enum merge_mode merge;
         enum xkb_map_flags mapFlags;
+        xkb_keysym_t    keysym;
         ParseCommon     *any;
         ExprDef         *expr;
         VarDef          *var;
@@ -208,41 +206,14 @@ typedef union YYSTYPE
         void            *geom;
         XkbFile         *file;
 
-
-/* Line 2077 of yacc.c  */
-#line 214 "src/xkbcomp/parser.h"
-} YYSTYPE;
+#line 210 "src/xkbcomp/parser.h" /* yacc.c:1909  */
+};
 # define YYSTYPE_IS_TRIVIAL 1
-# define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
 #endif
 
-#if ! defined YYLTYPE && ! defined YYLTYPE_IS_DECLARED
-typedef struct YYLTYPE
-{
-  int first_line;
-  int first_column;
-  int last_line;
-  int last_column;
-} YYLTYPE;
-# define yyltype YYLTYPE /* obsolescent; will be withdrawn */
-# define YYLTYPE_IS_DECLARED 1
-# define YYLTYPE_IS_TRIVIAL 1
-#endif
 
 
-#ifdef YYPARSE_PARAM
-#if defined __STDC__ || defined __cplusplus
-int _xkbcommon_parse (void *YYPARSE_PARAM);
-#else
-int _xkbcommon_parse ();
-#endif
-#else /* ! YYPARSE_PARAM */
-#if defined __STDC__ || defined __cplusplus
 int _xkbcommon_parse (struct parser_param *param);
-#else
-int _xkbcommon_parse ();
-#endif
-#endif /* ! YYPARSE_PARAM */
 
 #endif /* !YY__XKBCOMMON_SRC_XKBCOMP_PARSER_H_INCLUDED  */
